/*
 * parmes.cpp
 *
 *  Created on: 2010/09/21
 *      Author: shu
 */

#include "aligner_parameters.h"
#include "sequence_type.h"
#include "protein_type.h"
#include "dna_type.h"
#include "encoder.h"
#include "score_matrix.h"
#include "score_matrix_reader.h"
#include "logger.h"
#include <stdexcept>
#include <string>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <typeinfo>
#include <getopt.h>

using namespace std;

// dna
const uint32_t AlignerParameters::kSearchLengthDna = 24;
const int AlignerParameters::kThresholdDna = 18;
const int AlignerParameters::kSearchCutoffDna = 9;
const int AlignerParameters::kOpenGapDna = 5;
const int AlignerParameters::kExtendGapDna = 2;
const float AlignerParameters::kUngappedExtendNormalizedCutoffDna = 20.0;
const float AlignerParameters::kGappedExtendNormalizedCutoffDna = 30.0;
const float AlignerParameters::kFinalGappedExtendNormalizedCutoffDna = 100.0;
const float AlignerParameters::kGappedExtendNormalizedTriggerDna = 25.0;

//prot
const uint32_t AlignerParameters::kSearchLengthProt = 10;
const int AlignerParameters::kThresholdProt = 30;
const int AlignerParameters::kSearchCutoffProt = 4;
const int AlignerParameters::kOpenGapProt = 11;
const int AlignerParameters::kExtendGapProt = 1;
const float AlignerParameters::kUngappedExtendNormalizedCutoffProt = 7.0;
const float AlignerParameters::kGappedExtendNormalizedCutoffProt = 15.0;
const float AlignerParameters::kFinalGappedExtendNormalizedCutoffProt = 25.0;
const float AlignerParameters::kGappedExtendNormalizedTriggerProt = 22.0;

// common
const int AlignerParameters::kNumberThreads = 1;
const uint32_t AlignerParameters::kMaxNumberHits = 1 << 26;
const uint32_t AlignerParameters::kMaxQueriesLength = 1 << 27;
const uint32_t AlignerParameters::kMaxNumberAlignments = 10;
const uint32_t AlignerParameters::kMaxNumberOneSubjectAlignments = 1;
const bool AlignerParameters::kFilter = true;

AlignerParameters::AlignerParameters() :
		number_threads_(1), queries_filename_(""), db_filename_(""), output_filename_(
				""), query_type_(new ProteinType()), aligned_type_(
				new ProteinType()), search_length_(8), threshold_(30), search_cutoff_(
				1), max_number_hits_(1 << 21), max_queries_length_(1 << 27), max_number_alignments_(
				10), max_number_one_subject_alignments_(1), open_gap_(11), extend_gap_(
				1), ungapped_extend_normalized_cutoff_(7.0), gapped_extend_normalized_cutoff_(
				15.0), final_gapped_extend_normalized_cutoff_(25.0), gapped_extend_normalized_trigger_(
				22.0), kNextQueryIdIncrement_(1), kNextAlignmentIdIncrement_(1), filter_(
				true) {
	SetScoringParameters(*aligned_type_, string(""), open_gap_, extend_gap_);
}

AlignerParameters::~AlignerParameters() {
}

void AlignerParameters::SetParameters(int argc, char* argv[]) {
	string score_matrix_name = "";
	int c;
	Options options;
	SetDefaultOptions(options);
	extern char *optarg;
	extern int optind;
	optind = 1;

	while ((c = getopt(argc, argv, "a:b:d:E:F:G:i:l:L:M:o:q:r:s:S:t:T:v:"))
			!= -1) {
		switch (c) {
		case 'a':
			options.number_threads = atoi(optarg);
			break;

		case 'b':
			options.max_number_alignments = atoi(optarg);
			break;

		case 'd':
			options.db_filename = optarg;
			break;

		case 'E':
			options.extend_gap = atoi(optarg);
			break;

		case 'F':
			if (optarg[0] == 'T') {
				options.filter = true;
			} else if (optarg[0] == 'F') {
				options.filter = false;
			} else {
				throw invalid_argument(string("invalid option, -F is T or F."));
			}
			break;

		case 'G':
			options.open_gap = atoi(optarg);
			break;

		case 'i':
			options.queries_filename = optarg;
			break;

		case 'l':
			options.max_queries_length = atoi(optarg);
			break;

		case 'L':
			options.max_number_hits = atoi(optarg);
			break;

		case 'M':
			options.score_matrix_name = optarg;
			break;

		case 'o':
			options.output_filename = optarg;
			break;

		case 'q':
			if (optarg[0] == 'p') {
				options.query_type = SequenceTypePtr(new ProteinType());
			} else if (optarg[0] == 'd') {
				options.query_type = SequenceTypePtr(new DnaType());
			} else {
				throw invalid_argument(string("invalid option, -q is p or d."));
			}
			break;

		case 's':
			options.search_cutoff = atoi(optarg);
			break;

		case 'S':
			options.search_length = atoi(optarg);
			break;

		case 't':
			if (optarg[0] == 'p') {
				options.aligned_type = SequenceTypePtr(new ProteinType());
			} else if (optarg[0] == 'd') {
				options.aligned_type = SequenceTypePtr(new DnaType());
			} else {
				throw invalid_argument(string("invalid option, -t is p or d."));
			}
			break;

		case 'T':
			options.threshold = atoi(optarg);
			break;

		case 'v':
			options.max_number_one_subject_alignments = atoi(optarg);
			break;

		default:
			throw invalid_argument(
					"\nTry `ghostx --help' for more information.");
		}
	}
	SetParameters(options);
}

void AlignerParameters::SetDefaultOptions(Options &options) {
	options.number_threads = 1;
	options.queries_filename = "";
	options.db_filename = "";
	options.output_filename = "";
	options.score_matrix_name = "";
	options.query_type = SequenceTypePtr(new ProteinType());
	options.aligned_type = SequenceTypePtr(new ProteinType());
	options.search_length = 0;
	options.threshold = -1;
	options.search_cutoff = -1;
	options.max_number_hits = 0;
	options.max_queries_length = 0;
	options.max_number_alignments = 0;
	options.max_number_one_subject_alignments = 0;
	options.max_gapped_size = -1;
	options.open_gap = -1;
	options.extend_gap = -1;
	options.filter = kFilter;
	options.device_id = kCpuId;
}
void AlignerParameters::SetParameters(const Options &options) {
	// set default paraeters
	number_threads_ = kNumberThreads;
	max_number_hits_ = kMaxNumberHits;
	max_queries_length_ = kMaxQueriesLength;
	max_number_alignments_ = kMaxNumberAlignments;
	max_number_one_subject_alignments_ = kMaxNumberOneSubjectAlignments;
	filter_ = kFilter;

	if (typeid(*(options.query_type)) == typeid(DnaType)) {
		query_type_ = SequenceTypePtr(new DnaType());
	} else {
		query_type_ = SequenceTypePtr(new ProteinType());
	}

	if (typeid(*(options.aligned_type)) == typeid(DnaType)) {
		aligned_type_ = SequenceTypePtr(new DnaType());
		threshold_ = kThresholdDna;
		search_length_ = kSearchLengthDna;
		search_cutoff_ = kSearchCutoffDna;
		open_gap_ = kOpenGapDna;
		extend_gap_ = kExtendGapDna;
		ungapped_extend_normalized_cutoff_ = kUngappedExtendNormalizedCutoffDna;
		gapped_extend_normalized_cutoff_ = kGappedExtendNormalizedCutoffDna;
		final_gapped_extend_normalized_cutoff_ =
				kFinalGappedExtendNormalizedCutoffDna;
		gapped_extend_normalized_trigger_ = kGappedExtendNormalizedTriggerDna;
	} else {
		aligned_type_ = SequenceTypePtr(new ProteinType());
		threshold_ = kThresholdProt;
		search_length_ = kSearchLengthProt;
		search_cutoff_ = kSearchCutoffProt;
		open_gap_ = kOpenGapProt;
		extend_gap_ = kExtendGapProt;
		ungapped_extend_normalized_cutoff_ =
				kUngappedExtendNormalizedCutoffProt;
		gapped_extend_normalized_cutoff_ = kGappedExtendNormalizedCutoffProt;
		final_gapped_extend_normalized_cutoff_ =
				kFinalGappedExtendNormalizedCutoffProt;
		gapped_extend_normalized_trigger_ = kGappedExtendNormalizedTriggerProt;
	}

	// set parameters
	queries_filename_ = options.queries_filename;
	db_filename_ = options.db_filename;
	output_filename_ = options.output_filename;

	if (options.max_number_hits > 0) {
		max_number_hits_ = options.max_number_hits;
	}

	if (options.max_queries_length > 0) {
		max_queries_length_ = options.max_queries_length;
	}

	if (options.max_number_alignments > 0) {
		max_number_alignments_ = options.max_number_alignments;
	}

	if (options.max_number_one_subject_alignments > 0) {
		max_number_one_subject_alignments_ =
				options.max_number_one_subject_alignments;
	}

	if (options.threshold > -1) {
		threshold_ = options.threshold;
	}
	if (options.search_length > 0) {
		search_length_ = options.search_length;
	}
	if (options.search_cutoff > -1) {
		search_cutoff_ = options.search_cutoff;
	}
	if (options.open_gap > -1) {
		open_gap_ = options.open_gap;
	}
	if (options.extend_gap > -1) {
		extend_gap_ = options.extend_gap;
	}
	if (options.number_threads > 0) {
		number_threads_ = options.number_threads;
	}
	filter_ = options.filter;

	SetScoringParameters(*aligned_type_, options.score_matrix_name, open_gap_,
			extend_gap_);
}

void AlignerParameters::SetScoringParameters(SequenceType &type, string name,
		int open_gap, int extend_gap) {
	Encoder encoder(type);
	ScoreMatrixReader score_matrix_reader(encoder);
	if (name.length() != 0) {
		score_matrix_ = score_matrix_reader.Read(name);
	} else {
		score_matrix_ = score_matrix_reader.ReadDefault();
	}

	Statistics statistics(type);
	statistics.CalculateUngappedIdealKarlinParameters(score_matrix_,
			&ungapped_ideal_karlin_parameters_);
	ungapped_ideal_extend_nominal_cutoff_ =
			Statistics::NormalizedCutoff2NominalCutoff(
					ungapped_extend_normalized_cutoff_,
					ungapped_ideal_karlin_parameters_);
	gapped_ideal_extend_nominal_trigger_ = Statistics::Normalized2Nominal(
			gapped_extend_normalized_trigger_,
			ungapped_ideal_karlin_parameters_);
	statistics.CalculateGappedKarlinParameters(score_matrix_, open_gap,
			extend_gap, &gapped_karlin_parameters_);
	gapped_ideal_extend_nominal_cutoff_ =
			Statistics::NormalizedCutoff2NominalCutoff(
					gapped_extend_normalized_cutoff_,
					gapped_karlin_parameters_);
	final_gapped_ideal_extend_nominal_cutoff_ =
			Statistics::NormalizedCutoff2NominalCutoff(
					final_gapped_extend_normalized_cutoff_,
					gapped_karlin_parameters_);
	statistics.CalculateAlphaBeta(score_matrix_, open_gap, extend_gap, &alpha_,
			&beta_);
}
